// Copyright (C) 2010, 2011 Zeno Gantner
//
// This file is part of MyMediaLite.
//
// MyMediaLite is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MyMediaLite is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with MyMediaLite.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Globalization;
using System.IO;
using System.Linq;
using MyMediaLite.Data;
using MyMediaLite.DataType;
using MyMediaLite.Eval;
using MyMediaLite.IO;

namespace MyMediaLite.RatingPrediction
{
	/// <summary>baseline method for rating prediction</summary>
	/// <remarks>
	/// Uses the average rating value, plus a regularized user and item bias
	/// for prediction.
	///
	/// The method is described in section 2.1 of
	/// Yehuda Koren: Factor in the Neighbors: Scalable and Accurate Collaborative Filtering,
	/// Transactions on Knowledge Discovery from Data (TKDD), 2009.
	///
	/// One difference is that we support several iterations of alternating optimization,
	/// instead of just one.
	///
	/// This recommender supports incremental updates.
	/// </remarks>
	public class UserItemBaseline : IncrementalRatingPredictor, IIterativeModel
	{
		/// <summary>Regularization parameter for the user biases</summary>
		public double RegU { get; set; }

		/// <summary>Regularization parameter for the item biases</summary>
		public double RegI { get; set; }

		///
		public uint NumIter { get; set; }

		/// <summary>the global rating average</summary>
		protected double global_average;

		/// <summary>the user biases</summary>
		protected double[] user_biases;

		/// <summary>the item biases</summary>
		protected double[] item_biases;

		/// <summary>Default constructor</summary>
		public UserItemBaseline() : base()
		{
			RegU = 15;
			RegI = 10;
			NumIter = 10;
		}

		///
		public override void Train()
		{
			user_biases = new double[MaxUserID + 1];
			item_biases = new double[MaxItemID + 1];

			global_average = ratings.Average;

			for (uint i = 0; i < NumIter; i++)
				Iterate();
		}

		///
		public void Iterate()
		{
			OptimizeItemBiases();
			OptimizeUserBiases();
		}

		void OptimizeUserBiases()
		{
			int[] user_ratings_count = new int[MaxUserID + 1];

			for (int index = 0; index < ratings.Count; index++)
			{
				user_biases[ratings.Users[index]] += ratings[index] - global_average - item_biases[ratings.Items[index]];
				user_ratings_count[ratings.Users[index]]++;
			}
			for (int u = 0; u < user_biases.Length; u++)
				if (user_ratings_count[u] != 0)
					user_biases[u] = user_biases[u] / (RegU + user_ratings_count[u]);
		}

		void OptimizeItemBiases()
		{
			int[] item_ratings_count = new int[MaxItemID + 1];

			for (int index = 0; index < ratings.Count; index++)
			{
				item_biases[ratings.Items[index]] += ratings[index] - global_average - user_biases[ratings.Users[index]];
				item_ratings_count[ratings.Items[index]]++;
			}
			for (int i = 0; i < item_biases.Length; i++)
				if (item_ratings_count[i] != 0)
					item_biases[i] = item_biases[i] / (RegI + item_ratings_count[i]);
		}

		///
		public override double Predict(int user_id, int item_id)
		{
			double user_bias = (user_id < user_biases.Length && user_id >= 0) ? user_biases[user_id] : 0;
			double item_bias = (item_id < item_biases.Length && item_id >= 0) ? item_biases[item_id] : 0;
			double result = global_average + user_bias + item_bias;

			if (result > MaxRating)
				return MaxRating;
			if (result < MinRating)
				return MinRating;
			return result;
		}

		///
		public virtual void RetrainUser(int user_id)
		{
			if (UpdateUsers)
			{
				foreach (int index in ratings.ByUser[user_id])
					user_biases[user_id] += ratings[index] - global_average - item_biases[ratings.Items[index]];
				if (ratings.ByUser[user_id].Count != 0)
					user_biases[user_id] = user_biases[user_id] / (RegU + ratings.ByUser[user_id].Count);
			}
		}

		///
		public virtual void RetrainItem(int item_id)
		{
			if (UpdateItems)
			{
				foreach (int index in ratings.ByItem[item_id])
					item_biases[item_id] += ratings[index] - global_average;
				if (ratings.ByItem[item_id].Count != 0)
					item_biases[item_id] = item_biases[item_id] / (RegI + ratings.ByItem[item_id].Count);
			}
		}

		///
		public override void AddRating(int user_id, int item_id, double rating)
		{
			base.AddRating(user_id, item_id, rating);
			RetrainItem(item_id);
			RetrainUser(user_id);
		}

		///
		public override void UpdateRating(int user_id, int item_id, double rating)
		{
			base.UpdateRating(user_id, item_id, rating);
			RetrainItem(item_id);
			RetrainUser(user_id);
		}

		///
		public override void RemoveRating(int user_id, int item_id)
		{
			base.RemoveRating(user_id, item_id);
			RetrainItem(item_id);
			RetrainUser(user_id);
		}

		///
		protected override void AddUser(int user_id)
		{
			base.AddUser(user_id);

			double[] user_biases = new double[this.MaxUserID + 1];
			Array.Copy(this.user_biases, user_biases, this.user_biases.Length);
			this.user_biases = user_biases;
		}

		///
		protected override void AddItem(int item_id)
		{
			base.AddItem(item_id);

			double[] item_biases = new double[this.MaxItemID + 1];
			Array.Copy(this.item_biases, item_biases, this.item_biases.Length);
			this.item_biases = item_biases;
		}

		///
		public override void SaveModel(string filename)
		{
			using ( StreamWriter writer = Model.GetWriter(filename, this.GetType()) )
			{
				writer.WriteLine(global_average.ToString(CultureInfo.InvariantCulture));
				writer.WriteVector(user_biases);
				writer.WriteVector(item_biases);
			}
		}

		///
		public override void LoadModel(string filename)
		{
			using ( StreamReader reader = Model.GetReader(filename, this.GetType()) )
			{
				var global_average = double.Parse(reader.ReadLine(), CultureInfo.InvariantCulture);
				var user_biases = reader.ReadVector();
				var item_biases = reader.ReadVector();

				this.global_average = global_average;
				this.user_biases = user_biases.ToArray();
				this.item_biases = item_biases.ToArray();
			}
		}

		///
		public double ComputeFit()
		{
			return
				this.Evaluate(ratings)["RMSE"]
				+ RegU * Math.Pow(user_biases.EuclideanNorm(), 2)
				+ RegI * Math.Pow(item_biases.EuclideanNorm(), 2);
		}

		///
		public override string ToString()
		{
			return string.Format(
				CultureInfo.InvariantCulture,
				"{0} reg_u={1} reg_i={2} num_iter={3}",
				this.GetType().Name, RegU, RegI, NumIter);
		}
	}
}
